/*
  PersistentDestructibles.cpp - Destroyed solars remain destroyed.

  Jason Hood, 3 November, 2010.
  1 April, 2023: also destroy after FLPACKET_SERVER_SYSTEM_SWITCH_IN.

  When a solar is destroyed, the server sends a packet to the client saying so.
  When the system is revisited, the solar reappears, but is invincible (hits
  are registered, but because the server destroyed it, no damage is sent back).
  This plugin recognises destroyed solars and destroys them again.  It does so
  by making a list of the solars loaded and comparing that to the objects given
  a reputation by FLPACKET_SERVER_SETREPUTATION, which isn't sent for destroyed
  solars.  Currently only for MP; SP will require a slightly different approach.
  Destroyed objects are also removed from the Nav Map (including in SP), but
  this only works for "visit = 0"; "visit = 1" does something different, so the
  objects remain on the map.

  Copy PersistentDestructibles.dll to your EXE directory and add it to the
  [Libraries] section of dacom.ini.
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <set>

#define NAKED	__declspec( naked )
#define STDCALL __stdcall

DWORD dummy;
#define ProtectX( addr, size ) \
  VirtualProtect( addr, size, PAGE_EXECUTE_READWRITE, &dummy )
#define ProtectW( addr, size ) \
  VirtualProtect( addr, size, PAGE_READWRITE, &dummy )

#define RELOFS( from, to ) \
  *(DWORD*)(from) = (DWORD)(to) - (DWORD)(from) - 4

#define CALL( from, to ) \
  ((PBYTE)(from))[0] = 0xE8; \
  RELOFS( (PBYTE)(from)+1, to )

#define JMP( from, to ) \
  ((PBYTE)(from))[0] = 0xE9; \
  RELOFS( (PBYTE)(from)+1, to )

#define NEWABS( from, to, prev ) \
  prev = *(PDWORD)(from); \
  *(PDWORD)(from) = (DWORD)(to)


typedef std::set<UINT> SolarSet;
typedef SolarSet::const_iterator SolarSetCIter;

SolarSet solars;
int	 loading;
DWORD	 SetRep_Org, Launch_Org, SwitchIn_Org, Server_Org;
DWORD	 ret_addr, ecx_save;

struct ObjectDestroy
{
  UINT object;
  int  type;
};

typedef void (STDCALL* TDestroyObject)( int, const ObjectDestroy& );
TDestroyObject DestroyObject = (TDestroyObject)0x53bc80;


void Add_Object( UINT obj )
{
  solars.insert( obj );
}


void Rem_Object()
{
  solars.erase( *(PUINT)(ecx_save+0x58) );
}


void Destroy_Objects()
{
  loading = 0;

  ObjectDestroy od;
  od.type = 0;				// just disappears, 1 = explodes
  SolarSetCIter iter, end = solars.end();
  for (iter = solars.begin(); iter != end; ++iter)
  {
    od.object = *iter;
    DestroyObject( 1, od );
  }

  solars.clear();
}


// Create a set of objects read from the system file.
NAKED
void ReadObj_Hook()
{
  __asm {
	cmp	loading, -1
	je	done
	mov	loading, 1
	push	edx
	push	ecx
	push	[esp+0x20+12+12]
	call	Add_Object
	pop	ecx
	pop	ecx
	pop	edx
  done:
	cmp	ecx, ebx
	ret	12
  }
}


// Remove the objects seen by FLPACKET_SERVER_SETREPUTATION.
NAKED
void SetRep_Hook()
{
  __asm {
	cmp	loading, 0
	je	done
	pop	ret_addr
	mov	ecx_save, ecx
	call	SetRep_Org
	push	eax
	call	Rem_Object
	pop	eax
	jmp	ret_addr
  done:
	jmp	SetRep_Org
	align	16
  }
}


// FLPACKET_SERVER_LAUNCH indicates all objects have been processed - compare
// what was read with what was seen.
NAKED
void Launch_Hook()
{
  __asm {
	cmp	loading, 0
	je	done
	push	ecx
	call	Destroy_Objects
	pop	ecx
  done:
	jmp	Launch_Org
	align	16
  }
}


// Ditto for FLPACKET_SERVER_SYSTEM_SWITCH_IN.
NAKED
void SwitchIn_Hook()
{
  __asm {
	cmp	loading, 0
	je	done
	push	ecx
	call	Destroy_Objects
	pop	ecx
  done:
	jmp	SwitchIn_Org
	align	16
  }
}


// Remove a destroyed object from the Nav Map.
NAKED
void Destroy_Hook()
{
  __asm {
	push	eax
	mov	ecx, esp
	push	edi
	push	ecx
	mov	eax, 0x432240
	mov	ecx, 0x673320
	call	eax
	pop	eax
	cmp	eax, ds:[0x673324]
	je	done
	mov	byte ptr [eax+0x10], 0x80	// visit = 128
  done:
	pop	edi
	pop	esi
	ret	8
  }
}


void PatchServer()
{
  #define ADDR_SETREP	(0x6c0bbc4 - 0x6bd0000)
  #define ADDR_SWITCHIN (0x6c0c960 - 0x6bd0000)
  #define ADDR_LAUNCH	(0x6c0ca00 - 0x6bd0000)

  loading = -1;
  PBYTE server = (PBYTE)GetModuleHandle( "remoteserver.dll" );
  if (server)
  {
    ProtectW( server+ADDR_SETREP, 4 );
    ProtectW( server+ADDR_LAUNCH, 4 );
    NEWABS( server+ADDR_SETREP, SetRep_Hook, SetRep_Org );
    NEWABS( server+ADDR_SWITCHIN, SwitchIn_Hook, SwitchIn_Org );
    NEWABS( server+ADDR_LAUNCH, Launch_Hook, Launch_Org );
    loading = 0;
  }
}


NAKED
void Server_Hook()
{
  __asm {
	call	PatchServer
	jmp	Server_Org
	align	16
  }
}

DWORD Server_Addr = (DWORD)Server_Hook;


void Patch( void )
{
  #define ADDR_SERVER	((PBYTE)0x5a8119+2)
  #define ADDR_READOBJ	((PBYTE)0x548b4e)
  #define ADDR_DESTROY	((PBYTE)0x53bd2e)

  ProtectX( ADDR_DESTROY, 5 );
  ProtectX( ADDR_READOBJ, 5 );
  ProtectX( ADDR_SERVER,  4 );

  // Remoteserver is loaded every time a server is selected.
  NEWABS( ADDR_SERVER,	&Server_Addr, Server_Org );
  Server_Org = *(PDWORD)Server_Org;

  CALL( ADDR_READOBJ, ReadObj_Hook );
  JMP(	ADDR_DESTROY, Destroy_Hook );	// this part also works in SP
}


BOOL WINAPI DllMain( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved )
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    Patch();

  return TRUE;
}
